#include "Land.h"

/* The Constructor generates the terrain upon creation of a new instance */

Land::Land()
{
  generateLand();
  garaud = true;
}

Land::~Land()
{}

/* This function generates a random terrain and calculates
 * the normal vectors for shading.  It stores two versions,
 * one with flat shading, the other with Gouraud shading.
 * It then saves them into display lists to be quickly drawn.
 */

void Land::generateLand()
{
  int i,j;
  srand ((unsigned int)time(NULL));

  /* Create a 2-dimensional array of SIZE+1xSIZE+1 values,
   * each one corresponding to the height of the vertex at
   * each face.
   */
  for (i=0;i<SIZE+1;i++)
    for (j=0;j<SIZE+1;j++)
      landarray[i][j] = (VARIATION*(GLfloat)rand()/RAND_MAX) - (VARIATION/2);

  /* Set the material of the terrain to determine how it
   * responds to lighting. Since it is terrain, it is not
   * very shiny.
   */
  float white[] = {1,1,1,1};
  float black[] = {0,0,0,1};
  float shinyvec[] = {0};
  glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shinyvec);
  glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
  glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,black);

  /* Begin recording information for the first display list. */

  land = glGenLists(1);
  glNewList(land,GL_COMPILE);

  /* Brown color. */
  glColor3f(0.55f,0.27f,0.09f);

  /* Loop over the vertices and create a quadrilateral for each one. */
  for (i=0;i<SIZE;i++)
    for (j=0;j<SIZE;j++)
    {
      glBegin(GL_QUADS);

      /* This is calculated by taking the cross product of two
       * vectors pointing out from the pixel (i,j), resulting
       * in a normal vector.
       */
      glNormal3f(landarray[i][j]-landarray[i+1][j],landarray[i][j]-landarray[i][j+1],1.f);

      /* Vertices. */
      glVertex3f((GLfloat)i-SIZE/2,(GLfloat)j-SIZE/2,landarray[i][j]);
      glVertex3f((GLfloat)i-SIZE/2+1,(GLfloat)j-SIZE/2,landarray[i+1][j]);
      glVertex3f((GLfloat)i-SIZE/2+1,(GLfloat)j-SIZE/2+1,landarray[i+1][j+1]);
      glVertex3f((GLfloat)i-SIZE/2,(GLfloat)j-SIZE/2+1,landarray[i][j+1]);
      glEnd();
    }
  glEndList();

  /* Calculate the normal for each vertex (not each face)
   * for Gouraud Shading.
   */
  calculateGaraud();

  /* These calls do the same as above. */
  landg = glGenLists(1);
  glNewList(landg,GL_COMPILE);

  glColor3f(0.55f,0.27f,0.09f);
  
  for (i=0;i<SIZE;i++)
    for (j=0;j<SIZE;j++)
    {
      glBegin(GL_QUADS);

      /* Use the normal vectors calculated from our helper function for
       * each vertex.  This creates the smooth effect in our shading.
       */
      glNormal3f(gnormals[i][j][0],gnormals[i][j][1],1.f);
      glVertex3f((GLfloat)i-SIZE/2,(GLfloat)j-SIZE/2,landarray[i][j]);
      glNormal3f(gnormals[i+1][j][0],gnormals[i+1][j][1],1.f);
      glVertex3f((GLfloat)i-SIZE/2+1,(GLfloat)j-SIZE/2,landarray[i+1][j]);
      glNormal3f(gnormals[i+1][j+1][0],gnormals[i+1][j+1][1],1.f);
      glVertex3f((GLfloat)i-SIZE/2+1,(GLfloat)j-SIZE/2+1,landarray[i+1][j+1]);
      glNormal3f(gnormals[i][j+1][0],gnormals[i][j+1][1],1.f);
      glVertex3f((GLfloat)i-SIZE/2,(GLfloat)j-SIZE/2+1,landarray[i][j+1]);
      glEnd();
    }
  glEndList();
}


/* Since our generation functions did all the hard work,
 * all we have to do is call the call list to draw the land.
 */
void Land::drawLand()
{
  glPushMatrix();
  glTranslated(0,0,-3);
  if (garaud)
    glCallList(landg);
  else
    glCallList(land);
  glPopMatrix();
}

/* void Land::calculateGaraud() 
 *
 * This function calculates the appropriate normal vector for
 * each vertex. The function looks simple, but the math is rather
 * complex.
 *
 * Each vertex has four faces surrounding it, so in order to get
 * a smooth shading, all four sides must be taken into account, and
 * each face should use the same normal for this vertex. So we
 * compute the average of the four normal vectors of the four surrounding
 * faces, and store that average to be used each time this vertex is drawn.
 *
 * This is done like this:
 *
 *  |---------z1---------|       Point z3 is where our vertex is, (i,j). We want the average normal of
 *  |    A    |    B     |       A,B,C,D. We get this by taking the cross product of the two vectors from
 *  |         |          |       z3 that correspond to that face, then adding them up and dividing by four.
 *  z2--------z3-------z4|
 *  |    C    |    D     |       ((z1-z3)X(z2-z3) + (z4-z3)X(z1-z3) + (z5-z3)X(z4-z3) + (z2-z3)X(z5-z3))/4
 *  |         |          |
 *  |---------z5---------|       These are 3-dimensional vectors, so the values of i and j must also be considered,
 *                               but the i and j are always canceled out, leaving a value of either -1, 0, or 1.
 *
 *
 * The final resulting vector is ((z1 - z3 - z5)/2, (z2 - z4)/2, 1).
 */
void Land::calculateGaraud()
{
  GLfloat z1,z2,z3,z4,z5;
  for (int i=0;i<SIZE+1;i++)
    for (int j=0;j<SIZE+1;j++)
    {
      z1 = z2 = z4 = z5 = 0;
      z3 = landarray[i][j];
      if (i!=0)
        z1 = landarray[i-1][j];
      if (i!=SIZE+1)
        z5 = landarray[i+1][j];
      if (j!=0)
        z2 = landarray[i][j-1];
      if (j!=SIZE+1)
        z4 = landarray[i][j+1];
      gnormals[i][j][0] = z1/2-z3/2-z5/2;
      gnormals[i][j][1] = z2/2-z4/2;
    }
}