/* SDL Test
 * by Kyle Smith, 2012
 *
 * This is just a simple program to test the different 
 * features of SDL in the context of an OpenGL window.
 *
 */

#include "SDL.h"
#include "SDL_opengl.h"
#include "SDL_audio.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "Land.h"
#include "Cube.h"
#include "Sphere.h"


/* The width and height of the OpenGL window. */
#define WIDTH  800
#define HEIGHT 600

#define LIGHTDISTANCE 3

#define Cos(th) cos((GLfloat)3.1415926/180*(th))
#define Sin(th) sin((GLfloat)3.1415926/180*(th))

/* Global Variable definitions. */

int camAngleV;
int camAngleH;
int lightAngle;
double distance;
bool moveLight;

Land *land;
Cube *cube;
Sphere *sphere;

GLuint v, f;

/* static void lighting()
 * Handles the light initialization and calculates its
 * position.  Also draws a sphere at the source location
 */
static void lighting()
{
  //  Translate intensity to color vectors
  float Ambient[]   = {0,0,0,1.0};
  float Diffuse[]   = {1,1,1,1.0};
  float Specular[]  = {0,0,0,1.0};
  float Position[] =  {LIGHTDISTANCE*Cos(lightAngle),LIGHTDISTANCE*Sin(lightAngle),3,1.0};
  //  OpenGL should normalize normal vectors
  glEnable(GL_NORMALIZE);
  //  Enable lighting
  glEnable(GL_LIGHTING);
  //  Location of viewer for specular calculations
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,0);
  //  glColor sets ambient and diffuse color materials
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  //  Enable light 0
  glEnable(GL_LIGHT0);
  //  Set ambient, diffuse, specular components and position of light 0
  glLightfv(GL_LIGHT0,GL_AMBIENT ,Ambient);
  glLightfv(GL_LIGHT0,GL_DIFFUSE ,Diffuse);
  glLightfv(GL_LIGHT0,GL_SPECULAR,Specular);
  glLightfv(GL_LIGHT0,GL_POSITION,Position);

  // The sphere is to be drawn in the same place as the light source
  sphere->setPos(Position[0],Position[1],Position[2]);
  sphere->draw();
  
}

/* static void repaint()
 * Draws everything on the screen. This function is called
 * every iteration of the main loop.
 */
static void repaint()
{  
  /* Clear the color plane and the z-buffer */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glPushMatrix();
  glLoadIdentity();

  /* Calculate the Euclidean coordinates of the camera position
   * based on the spherical coordinates */
  GLfloat camx, camy, camz;
  camx = (GLfloat)distance * Cos(camAngleV) * Cos(camAngleH);
  camy = (GLfloat)distance * Cos(camAngleV) * Sin(camAngleH);
  camz = (GLfloat)distance * Sin(camAngleV);

  gluLookAt(camx, camy, camz, 0,0,0,0,0,Cos(camAngleV));
  lighting();

  // Draws our land object
  land->drawLand();

  // Scale and draw the cube
  glPushMatrix();
  glScaled(2,2,2);
  cube->draw();
  glPopMatrix();


  // Draw the axes
  glScaled(10,10,10);
  glColor3f(0,0,0);

  glBegin(GL_LINES);
  glVertex3f(1,0,0);
  glVertex3f(-1,0,0);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(0,-1,0);
  glVertex3f(0,1,0);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(0,0,-1);
  glVertex3f(0,0,1);
  glEnd();

  glPopMatrix();

  /* finally, swap the back and front buffers */
  SDL_GL_SwapBuffers();
}

/* static void setup_sdl()
 * Initializes the necessary variables and sets up
 * SDL to handle OpenGL
 */
static void setup_sdl() 
{
  const SDL_VideoInfo* video;

  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
    exit(1);
  }

  /* Quit SDL properly on exit */
  atexit(SDL_Quit);

  /* Get the current video information */
  video = SDL_GetVideoInfo( );
  if( video == NULL ) {
    fprintf(stderr, "Couldn't get video information: %s\n", SDL_GetError());
    exit(1);
  }

  /* Set the minimum requirements for the OpenGL window */
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

  /* Note the SDL_DOUBLEBUF flag is not required to enable double 
  * buffering when setting an OpenGL video mode. 
  * Double buffering is enabled or disabled using the 
  * SDL_GL_DOUBLEBUFFER attribute.
  */
  if( SDL_SetVideoMode( WIDTH, HEIGHT, video->vfmt->BitsPerPixel, SDL_OPENGL ) == 0 ) {
    fprintf(stderr, "Couldn't set video mode: %s\n", SDL_GetError());
    exit(1);
  }
}

/* static void setup_opengl()
 * Initializes the state for OpenGL.
 */
static void setup_opengl()
{
  float aspect = (float)WIDTH / (float)HEIGHT;

  /* Make the viewport cover the whole window */
  glViewport(0, 0, WIDTH, HEIGHT);

  /* Set the camera projection matrix:
  * field of view: 90 degrees
  * near clipping plane at 0.1
  * far clipping plane at 100.0
  */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective(60.0, aspect, 0.1, 100.0);
  /* We're done with the camera, now matrix operations 
  * will affect the modelview matrix
  * */
  glMatrixMode(GL_MODELVIEW);

  /* set the clear color to grey */
  glClearColor(0.5f , 0.5f ,0.5f , 0);

  /* We want z-buffer tests enabled*/
  glEnable(GL_DEPTH_TEST);

  /* Do draw back-facing polygons*/
  glDisable(GL_CULL_FACE);
}

/* static void main_loop()
 * This will run until user input breaks out of the loop.
 * Accepts keyboard commands and exits on esc.
 */
static void main_loop() 
{
  SDL_Event event;
  Uint8 *keystate;
  while (1) {
    keystate = SDL_GetKeyState(NULL);

    /* process pending events */
    while( SDL_PollEvent( &event ) ) {
      switch( event.type ) {
      case SDL_KEYDOWN:
        switch ( event.key.keysym.sym ) {
        case SDLK_ESCAPE:
          exit(0);
          break;
        default:
          //no default key processing
          //(stops compiler warnings for unhandled SDL keydefs)
          break;
        case SDLK_g:
          land->garaud = !land->garaud;
          break;
        case SDLK_m:
          moveLight = !moveLight;
        }
        break;

      case SDL_QUIT:
        exit (0);
        break;
      }
    }

    /* Process key inputs to change camera position.
     * These are outside the loop because holding the
     * key should sustain the effect, and to enable
     * simultaneous key inputs.
     */

    if (keystate[SDLK_RIGHT])
      camAngleH += 4;
    if (keystate[SDLK_LEFT])
      camAngleH -= 4;
    if (keystate[SDLK_UP])
      camAngleV += 4;
    if (keystate[SDLK_DOWN])
      camAngleV -= 4;
    if (keystate[SDLK_EQUALS])
      distance -= 0.3;
    if (keystate[SDLK_MINUS])
      distance += 0.3;

    if (distance < 0)
      distance = 0;

    // Mod the camera angle so it doesn't get too high.
    camAngleH = camAngleH % 360;
    camAngleV = camAngleV % 360;

    // Keep the light source moving.
    if (moveLight)
      lightAngle += 5;

    /* update the screen */    
    repaint();

    /* Wait 50ms to avoid using up all the CPU time */
    SDL_Delay( 50 );
  }
}

int main(int argc, char* argv[])
{
  setup_sdl();

  // Initialize state vars.
  camAngleH = 0;
  camAngleV = 45;
  distance = 20;
  lightAngle = 0;
  setup_opengl();

  // Create objects we want to draw.
  land = new Land();
  cube = new Cube(0,0,0);
  sphere = new Sphere(0,0,0,1,1,0.8f);

  moveLight = true;
  
  main_loop();

  return 0;
}